"""
Scrolling with the middle mouse button.

Useful for:
- smoother scrolling
- horizontal scrolling: the horizontal scrollbar is blocked by the
  half-transparent lower toolbar
"""

import logging

import openpaperwork_core
import openpaperwork_gtk.deps


GI_AVAILABLE = False
GTK_AVAILABLE = False

try:
    import gi
    GI_AVAILABLE = True
except (ImportError, ValueError):
    pass

if GI_AVAILABLE:
    try:
        gi.require_version('Gdk', '3.0')
        from gi.repository import Gdk
        GTK_AVAILABLE = True
    except (ImportError, ValueError):
        pass


LOGGER = logging.getLogger(__name__)


class AutoScrollingHandler(object):
    def __init__(self, core, scrollview):
        self.core = core
        self.refcount = 0
        self.mouse_start_position = (0, 0)  # relative to the screen
        self.mouse_position = (0, 0)  # relative to the screen

        self.cursors = {
            'inactive': None,
            'active': None,
        }

        try:
            display = scrollview.get_display()
            self.cursors['active'] = Gdk.Cursor.new_for_display(
                display, Gdk.CursorType.TCROSS
            )
        except TypeError:
            # may not work with pure-Wayland systems
            pass

        self.scrollview = scrollview
        scrollview.add_events(
            Gdk.EventMask.BUTTON_PRESS_MASK | Gdk.EventMask.POINTER_MOTION_MASK
        )
        self.button_press_id = scrollview.connect(
            "button-press-event", self._on_button_press
        )
        self.button_release_id = scrollview.connect(
            "button-release-event", self._on_button_release
        )
        self.motion_notify_id = scrollview.connect(
            "motion-notify-event", self._on_mouse_motion
        )

    def disable(self):
        self.scrollview.disconnect(self.button_press_id)
        self.scrollview.disconnect(self.button_release_id)
        self.scrollview.disconnect(self.motion_notify_id)

    def _on_tick(self):
        if self.mouse_start_position is None:
            self.refcount -= 1
            return

        hdiff = (self.mouse_position[0] - self.mouse_start_position[0]) / 5
        vdiff = (self.mouse_position[1] - self.mouse_start_position[1]) / 5

        hadj = self.scrollview.get_hadjustment()
        vadj = self.scrollview.get_vadjustment()

        hval = hadj.get_value() + hdiff
        hval = max(hval, hadj.get_lower())
        hval = min(hval, hadj.get_upper())
        hadj.set_value(hval)
        vval = vadj.get_value() + vdiff
        vval = max(vval, vadj.get_lower())
        vval = min(vval, vadj.get_upper())
        vadj.set_value(vval)

        self.core.call_one("mainloop_schedule", self._on_tick, delay_s=0.1)

    def _on_button_press(self, scrollview, event):
        if event.button != 2:
            return
        LOGGER.info("Starting autoscrolling")
        self.mouse_start_position = (event.x_root, event.y_root)
        self.mouse_position = (event.x_root, event.y_root)
        if self.refcount <= 0:
            self.core.call_one("mainloop_schedule", self._on_tick, delay_s=0.1)
        self.refcount += 1
        self.scrollview.get_window().set_cursor(self.cursors['active'])

    def _on_button_release(self, scrollview, event):
        if event.button != 2:
            return
        LOGGER.info("Ending autoscrolling")
        self.mouse_start_position = None
        self.scrollview.get_window().set_cursor(self.cursors['inactive'])

    def _on_mouse_motion(self, scrollview, event):
        if self.mouse_start_position is None:
            return
        self.mouse_position = (event.x_root, event.y_root)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_gesture_autoscrolling',
        ]

    def get_deps(self):
        return []

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def gesture_enable_autoscrolling(self, scrollview):
        LOGGER.info("Enabling autoscrolling on %s", scrollview)
        try:
            return AutoScrollingHandler(self.core, scrollview)
        except TypeError as exc:
            # may happen with Wayland
            LOGGER.error("Failed to switch mouse cursor", exc_info=exc)
            return None



import datetime
import logging

import openpaperwork_core
import openpaperwork_core.deps

from .. import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -100

    def __init__(self):
        super().__init__()
        self.url_selected = None
        self.method_radio = None
        self.method_set_file_urls_callbacks = {}
        self.windows = []
        self.widget_tree = None

    def get_interfaces(self):
        return [
            'gtk_bug_report_dialog',
            'gtk_window_listener',
            'screenshot_provider',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'external_apps',
                'defaults': [
                    'openpaperwork_core.external_apps.dbus',
                    'openpaperwork_core.external_apps.windows',
                    'openpaperwork_core.external_apps.xdg',
                ],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
            {
                'interface': 'screenshot',
                'defaults': ['openpaperwork_gtk.screenshots'],
            },
        ]

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def _shorten_url(self, url):
        if "://" not in url:
            return url
        dir_path = self.core.call_success("fs_dirname", url)
        dir_name = self.core.call_success("fs_basename", dir_path)
        if dir_name is None or dir_name == "":
            return url
        file_name = self.core.call_success("fs_basename", url)
        return dir_name + "/" + file_name

    def _format_date(self, date):
        if date is None:
            return _("Now")
        return date.strftime("%c")

    def _format_timestamp(self, date):
        if date is None:
            date = datetime.datetime.now()
        return int(date.timestamp())

    def _refresh_attachment_page_complete(self, widget_tree):
        # check we have at least one entry toggled
        model = widget_tree.get_object("bug_report_model")
        assistant = widget_tree.get_object("bug_report_dialog")
        page = widget_tree.get_object("bug_report_attachment_selector")

        enabled = False
        for row in model:
            if row[0]:
                enabled = True
            if row[0] and "://" not in row[4]:
                enabled = False
                break
        assistant.set_page_complete(page, enabled)

    def _i18n_file_size(self, file_size):
        if isinstance(file_size, str) or file_size <= 0:
            return ""
        return self.core.call_success("i18n_file_size", file_size)

    def open_bug_report(self):
        self.method_radio = None
        self.url_selected = None

        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "openpaperwork_gtk.bug_report", "bug_report.glade"
        )
        dialog = self.widget_tree.get_object("bug_report_dialog")

        self.core.call_all("bug_report_complete", self.widget_tree)

        self.widget_tree.get_object("bug_report_toggle_renderer").connect(
            "toggled", self._on_attachment_toggle, self.widget_tree
        )
        self.widget_tree.get_object("bug_report_treeview").connect(
            "row-activated", self._on_row_selected, self.widget_tree
        )
        self.widget_tree.get_object("bug_report_open_file").connect(
            "clicked", self._open_selected
        )

        model = self.widget_tree.get_object("bug_report_model")
        model.clear()

        inputs = {}
        self.core.call_all("bug_report_get_attachments", inputs)
        now = datetime.datetime.now()
        for (k, v) in inputs.items():
            v['id'] = k
            v['sort_date'] = (
                v['date'].timestamp()
                if v['date'] is not None
                else now.timestamp()
            )
        inputs = list(inputs.values())
        inputs.sort(
            key=lambda i: (-i['sort_date'], i['file_type'], i['file_url']),
        )
        for i in inputs:
            model.append(
                [
                    i['include_by_default'],
                    i['sort_date'],
                    self._format_date(i['date']),
                    i['file_type'],
                    i['file_url'],
                    self._shorten_url(i['file_url']),
                    i['file_size'],
                    self._i18n_file_size(i['file_size']),
                    i['id'],
                ]
            )

        for i in inputs:
            if not i['include_by_default']:
                continue
            self.core.call_all(
                "on_bug_report_attachment_selected", i['id'], self.widget_tree
            )

        if len(self.windows) > 0:
            dialog.set_transient_for(self.windows[-1])
        dialog.connect("close", self._on_close)
        dialog.connect("cancel", self._on_close)
        dialog.connect("prepare", self._on_prepare, self.widget_tree)

        dialog.set_visible(True)
        self.core.call_all("on_gtk_window_opened", dialog)
        self._refresh_attachment_page_complete(self.widget_tree)

    def close_bug_report(self):
        dialog = self.widget_tree.get_object("bug_report_dialog")
        dialog.destroy()

    def _on_attachment_toggle(self, cell_renderer, row_path, widget_tree):
        model = widget_tree.get_object("bug_report_model")
        val = not model[row_path][0]
        model[row_path][0] = val
        if val:
            self.core.call_all(
                "on_bug_report_attachment_selected",
                model[row_path][-1], widget_tree
            )
        else:
            self.core.call_all(
                "on_bug_report_attachment_unselected",
                model[row_path][-1], widget_tree
            )

        self._refresh_attachment_page_complete(widget_tree)

    def _on_row_selected(self, treeview, row_path, column, widget_tree):
        self._on_url_selected(widget_tree)

    def _on_url_selected(self, widget_tree):
        button = widget_tree.get_object("bug_report_open_file")
        treeview = widget_tree.get_object("bug_report_treeview")
        model = widget_tree.get_object("bug_report_model")
        (has_selected, model_iter) = (
            treeview.get_selection().get_selected()
        )
        if not has_selected or model_iter is None:
            button.set_sensitive(False)
            return
        self.url_selected = model[model_iter][4]
        button.set_sensitive("://" in self.url_selected)

    def _open_selected(self, button):
        self.core.call_success("external_app_open_file", self.url_selected)

    def bug_report_update_attachment(
            self, attachment_id, infos: dict, widget_tree):
        model = widget_tree.get_object("bug_report_model")
        for row in model:
            if row[-1] != attachment_id:
                continue
            if 'date' in infos:
                row[1] = self._format_timestamp(infos['date'])
                row[2] = self._format_date(infos['date'])
            if 'file_type' in infos:
                row[3] = infos['file_type']
            if 'file_url' in infos:
                row[4] = infos['file_url']
                row[5] = self._shorten_url(infos['file_url'])
            if 'file_size' in infos:
                row[6] = infos['file_size']
                row[7] = self._i18n_file_size(infos['file_size'])
            break
        self._on_url_selected(widget_tree)
        self._refresh_attachment_page_complete(widget_tree)

    def bug_report_get_attachment_file_url(self, attachment_id, widget_tree):
        model = widget_tree.get_object("bug_report_model")
        for row in model:
            if row[-1] != attachment_id:
                continue
            if "://" not in row[4]:
                return None
            return row[4]
        return None

    def _on_close(self, dialog):
        dialog.set_visible(False)
        self.core.call_all("on_gtk_window_closed", dialog)

    def bug_report_add_method(
            self, title, description,
            enable_callback, disable_callback,
            widget_tree):
        widget_tree_method = self.core.call_success(
            "gtk_load_widget_tree",
            "openpaperwork_gtk.bug_report", "bug_report_method.glade"
        )
        widget_tree_method.get_object("bug_report_method_title").set_text(
            title
        )
        widget_tree_method.get_object(
            "bug_report_method_description"
        ).set_markup(description)

        widget_tree.get_object("bug_report_methods").pack_start(
            widget_tree_method.get_object("bug_report_method"),
            expand=False, fill=False, padding=20
        )

        radio = widget_tree_method.get_object("bug_report_method_radio")
        radio.connect(
            "toggled", self._on_method_toggled,
            widget_tree, enable_callback, disable_callback
        )
        if self.method_radio is None:
            self.method_radio = radio
            enable_callback(widget_tree)
        else:
            radio.join_group(self.method_radio)
        self.method_radio.set_active(True)

    def _on_method_toggled(
            self, radio, widget_tree, enable_callback, disable_callback):
        if radio.get_active():
            enable_callback(widget_tree)
        else:
            disable_callback(widget_tree)

    def _on_prepare(self, assistant, page, widget_tree):
        self._set_file_urls(widget_tree)

    def _set_file_urls(self, widget_tree):
        model = widget_tree.get_object("bug_report_model")
        file_urls = [row[4] for row in model if row[0]]
        self.core.call_all("bug_report_set_file_urls_to_send", file_urls)

    def screenshot_snap_all_doc_widgets(self, out_dir):
        if self.widget_tree is None:
            return
        self.core.call_success(
            "screenshot_snap_widget",
            self.widget_tree.get_object("bug_report_dialog"),
            self.core.call_success("fs_join", out_dir, "bug_report.png")
        )



import datetime

import openpaperwork_core


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'gtk_calendar_popover',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
        ]

    def gtk_calendar_add_popover(self, gtk_entry):
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "openpaperwork_gtk.widgets.calendar", "calendar.glade"
        )
        widget_tree.get_object("calendar_popover").set_relative_to(
            gtk_entry
        )
        gtk_entry.connect(
            "icon-release", self._open_calendar, widget_tree
        )
        widget_tree.get_object("calendar_calendar").connect(
            "day-selected-double-click",
            self._update_date, gtk_entry, widget_tree
        )

    def _open_calendar(
            self, gtk_entry, icon_pos, event, widget_tree):
        date = self.core.call_success(
            "i18n_parse_date_short", gtk_entry.get_text()
        )
        if date is None:
            return
        widget_tree.get_object("calendar_calendar").select_month(
            date.month - 1, date.year
        )
        widget_tree.get_object("calendar_calendar").select_day(
            date.day
        )
        widget_tree.get_object("calendar_popover").set_visible(True)

    def _update_date(self, gtk_calendar, gtk_entry, widget_tree):
        date = widget_tree.get_object("calendar_calendar").get_date()
        date = datetime.datetime(year=date[0], month=date[1] + 1, day=date[2])
        date = self.core.call_success("i18n_date_short", date)
        gtk_entry.set_text(date)
        widget_tree.get_object("calendar_popover").set_visible(False)



import logging


GI_AVAILABLE = False
GTK_AVAILABLE = False


try:
    import gi
    from gi.repository import Gio
    from gi.repository import GLib
    GI_AVAILABLE = True
except (ValueError, ImportError):
    pass

if GI_AVAILABLE:
    try:
        gi.require_version('Gtk', '3.0')
        from gi.repository import Gtk
        GTK_AVAILABLE = True
    except (ImportError, ValueError):
        pass


import openpaperwork_core  # noqa: E402

from . import deps  # noqa: E402


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.activated = False
        self.app = None

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_init',
        ]

    def init(self, core):
        super().init(core)

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(deps.GTK)

    def gtk_init(self, main_func, *args):
        if hasattr(GLib, 'set_application_name'):
            GLib.set_application_name("Paperwork")
        GLib.set_prgname("work.openpaper.Paperwork")

        self.app = Gtk.Application(
            application_id="work.openpaper.Paperwork",
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE |
            Gio.ApplicationFlags.HANDLES_OPEN
        )
        self.app.connect("handle-local-options", main_func, *args)
        self.app.connect("activate", self._on_activate)
        self.app.connect("open", self._on_open)
        Gtk.Application.set_default(self.app)
        self.app.register()
        # self.app.run(in_args)
        self.app.run()

    def on_quit(self):
        if self.app is not None:
            self.app.quit()

    def gtk_get_app(self):
        return self.app

    # called when paperwork is launched without a file to import as argument.
    # when a second instance of paperwork is run, this is called in the main
    # instance instead.
    def _on_activate(self, _app):
        if not self.activated:
            self.activated = True
            self.core.call_all("on_gtk_initialized")

            LOGGER.info("Ready")
            self.core.call_one("mainloop", halt_on_uncaught_exception=False)
            LOGGER.info("Quitting")
            self.core.call_all("config_save")
            self.core.call_all("on_quit")
        else:
            # a second instance was activated, we are the main one
            self.core.call_all("mainwindow_present")

    # called when paperwork is launched with files to import as argument.
    # when a second instance of paperwork is run, this is called in the main
    # instance instead.
    def _on_open(self, app, files, _count, _hint):
        uris = [file.get_uri() for file in files]
        self.core.call_one(
            "mainloop_schedule",
            self.core.call_all, "gtk_doc_import", uris
        )
        # either start paperwork if we are the main instance, or focus it if
        # this signal is raised remotely
        self._on_activate(app)



import openpaperwork_core.mainloop.tests


class TestCallback(openpaperwork_core.mainloop.tests.AbstractTestCallback):
    def get_plugin_name(self):
        return "openpaperwork_gtk.mainloop.glib"


class TestPromise(openpaperwork_core.mainloop.tests.AbstractTestPromise):
    def get_plugin_name(self):
        return "openpaperwork_gtk.mainloop.glib"
